#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <math.h>
#include "hopfioctl.h"


unsigned char make_BCD (int i_wert)
{
  unsigned char b_BCD, b_upper, b_lower;

  i_wert %= 100;
  b_upper = i_wert / 10;
  b_lower = i_wert - b_upper * 10;
  b_BCD = (b_upper << 4) + b_lower;

  return (b_BCD);
}


unsigned long int bin_byte (unsigned char c_dec)
{
  int i;
  unsigned long int result = 0;

  for (i=7; i>=0; i--) {
    if (c_dec >> i) {
      result += (unsigned long int) pow (10, i);
      c_dec -= (1 << i);
    }
  }
  return (result);
}


int main (int argc, char *argv[])
{
  int i_fd;
  char ac_time[13];
  unsigned char b_status;
  unsigned char pc_time[8];
  int msec;
  time_t t_time;
  struct tm x_time;
  unsigned long int port;
  int guete;

  i_fd = open ("/dev/hopf6039", O_RDONLY);
  if (i_fd < 0) {
    perror ("Konnte Device nicht ffnen: ");
    exit (i_fd);
  }
  if (argc > 1) {
    if (!strcmp(argv[1], "help")) {
      printf ("\tDCF77 Testtool V0.1\n");
      printf ("\t\tby\n");
      printf ("Sven Bauhan <S.Bauhan@tu-bs.de> & Frank Haverkamp <haver@ibr.cs.tu-bs.de>\n");
      printf ("Usage:\n\tdcf [help | reset | slave | set | port <address>]\n");
      exit (0);
    }
    else if (!strcmp (argv[1], "reset")) {
      ioctl (i_fd, HOPF_IOCS_RESET);
      exit (0);
    }
    else if (!strcmp (argv[1], "slave")) {
      ioctl (i_fd, HOPF_IOCS_SLAVE);
      exit (0);
    }
    else if (!strcmp (argv[1], "master")) {
      ioctl (i_fd, HOPF_IOCS_MASTER);
      exit (0);
    }
    else if (!strcmp (argv[1], "set")) {
      /* Uhr stellen */
      pc_time[7] = 0x50;
      t_time = time (NULL);
      memcpy (&x_time, localtime (&t_time), sizeof (struct tm));
      pc_time[0] = make_BCD (x_time.tm_sec);
      pc_time[1] = make_BCD (x_time.tm_min);
      pc_time[2] = make_BCD (x_time.tm_hour);
      pc_time[3] = make_BCD (x_time.tm_wday);
      pc_time[4] = make_BCD (x_time.tm_yday);
      pc_time[5] = make_BCD (x_time.tm_mon);
      pc_time[6] = make_BCD (x_time.tm_year);
      ioctl (i_fd, HOPF_IOCS_SET, &pc_time[0]);
      exit (0);
    }
    else if (!strcmp (argv[1], "port")) {
      if (argc < 3) {
	printf ("Error: Too few arguments!\n");
	exit (1);
      }
      port = atol (argv[2]);
      printf ("Setting port address to 0x%x.\n", (unsigned int) port);
      if (ioctl (i_fd, HOPF_IOCS_PORT, port) < 0) {
	perror ("Fehler beim Aendern der Portadresse: ");
	exit (1);
      }
      exit (0);
    }
  }
  if (read (i_fd, ac_time, 12)) {
    perror ("Lesefehler: ");
  }
  else  printf ("Zeit: %s\n", ac_time);
  ioctl (i_fd, HOPF_IOCG_STATUS, &b_status);
  printf ("Statusbyte : %d = %ld binaer\n", b_status, bin_byte (b_status));
  guete = (int) ioctl (i_fd, HOPF_IOCQ_GUETE) / 2.55 + .5;
  printf ("Guete: %d%%\n", guete);

  msec = ioctl (i_fd, HOPF_IOCG_TIME, &pc_time);
  if (msec < 0) {
    perror ("Lesefehler: ");
  }
  else {
    printf ("%2d.%2d.%2d %2d:%2d:%2d:%3d\n", pc_time[3], pc_time[5],
	    pc_time[6], pc_time[2], pc_time[1], pc_time[0], msec);
  }
  close (i_fd);
  return 0;
}
