/******************************************************************************
 *
 * 1999
 * Sven Bauhan <S.Bauhan@tu-bs.de>
 * Frank Haverkamp <haver@ibr.cs.tu-bs.de>
 *
 * $Id: $
 *
 * $Log: $
 *
 *****************************************************************************/

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#define __NO_VERSION__      /* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

char kernel_version[] = UTS_RELEASE;

#include <linux/types.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/ioport.h>   /* check/request/release_region */
#include <linux/delay.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/ioctl.h>
#include <asm/segment.h>    /* verify_read, verify_write, get_user, put_user */
#include <asm/page.h>       /* PAGE_SIZE */
#include <asm/bitops.h>     /* set_bit, clear_bit, change_bit, test_bit */
#include <asm/current.h>    /* current pointer handling of new kernels */
#include <asm/uaccess.h>

#include "hopf.h"
#include "hopfioctl.h"

#if 0
#if defined(__alpha__) || (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
# define count_t unsigned long
# define read_write_t long
#else
# define count_t int
# define read_write_t int
#endif
#endif

/******************************************************************************
 *
 * Modulparameter
 *
 *****************************************************************************/

static unsigned int hopf_major = HOPF_MAJOR; /* device-driver major */
static unsigned long port_address = PORT_ADDRESS;
static unsigned char hopf_status;
static unsigned char hopf_time[HOPF_TIMESIZE];
    
/******************************************************************************
 *
 * This is the linux interface to the driver. It contains the functions the 
 * driver should provide to the user. The oder of the fields is important.
 * This structure is given to the kernel initialising the driver. It can be
 * changed afterwards in the open function of the driver to offer other 
 * functionalities. In this version this doesn't happen, but in later 
 * revistions we may consider to use this possibility.
 * 
 *****************************************************************************/


/******************************************************************************
 *
 * Accessing the IO-prots is made with: 
 * checkregion()
 * request_region()
 * release_region()
 *
 * Accessing ISA-memory under 1MB can easily performed with the functions:
 * unsigned readb(address)           void writeb(unsigned value, address)
 * unsigned readw(address)    and    void writew(unsigned value, address)
 * unsigned readl(address)           void writel(unsigned value, address)
 * memset_io(address, value, count)
 * memcpy_fromio(dest, source, nbytes)
 * memcpy_toio(dest, source, nbytes)
 *
 *****************************************************************************/


static unsigned long hopf_test (void)
{
    unsigned long int port_test, kenn_addr;

    for (port_test=SECTION_BEGIN; port_test<=SECTION_END;
	 port_test+=PORT_SIZE) {
        kenn_addr = port_test + KENNUNGS_OFFSET;
        if ((inb_p (kenn_addr) == 0x58) && (inb_p (kenn_addr+1) == 0x4E)) {
	    port_address = port_test;
	    return (port_address);
	}
    }
    port_address = PORT_ADDRESS;
    return (-ENODEV);
}


static int hopf_read_proc(char *buf, char **start, off_t offset,
			    int len, int unused)
{
    unsigned char wert, byte;
    unsigned int msec;

    len = 0;
    DPRINTK ("proc file called\n");
    /* --- printing date --- */
    /* get day of week */
    wert = inb_p (port_address + 4);
    switch (wert) {
    case 1:
        len += sprintf (buf + len, "Montag\n");
	break;
    case 2:
        len += sprintf (buf + len, "Dienstag\n");
	break;
    case 3:
        len += sprintf (buf + len, "Mittwoch\n");
	break;
    case 4:
        len += sprintf (buf + len, "Donnerstag\n");
	break;
    case 5:
        len += sprintf (buf + len, "Freitag\n");
	break;
    case 6:
        len += sprintf (buf + len, "Samstag\n");
	break;
    case 7:
        len += sprintf (buf + len, "Sonntag\n");
	break;
    }
    /* get day */
    wert = inb_p (port_address + 3);
    byte = wert >> 4;
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c.", byte + 48);
    /* get month */
    wert = inb_p (port_address + 5);
    byte = wert >> 4;
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c.", byte + 48);
    /* get year */
    wert = inb_p (port_address + 6);
    byte = wert >> 4;
    if (byte < 9)  len += sprintf (buf + len, "20");
    else  len += sprintf (buf + len, "19");
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c\n", byte + 48);
    /* --- printing time --- */
    /* get hour */
    wert = inb_p (port_address + 2);
    byte = wert >> 4;
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c:", byte + 48);
    /* get minutes */
    wert = inb_p (port_address + 1);
    byte = wert >> 4;
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c:", byte + 48);
    /* get seconds */
    wert = inb_p (port_address + 0);
    byte = wert >> 4;
    len += sprintf (buf + len, "%c", byte + 48);
    byte = wert & 0xf;
    len += sprintf (buf + len, "%c:", byte + 48);
    /* get millisconds */
    msec = inw (port_address + 0x38);
    byte = msec / 100;
    len += sprintf (buf + len, "%c", byte + 48);
    msec -= byte * 100;
    byte = msec / 10;
    len += sprintf (buf + len, "%c", byte + 48);
    msec -= byte * 10;
    len += sprintf (buf + len, "%c\n", byte + 48);
    /* --- printing status --- */
    wert = inb_p (port_address + 7);
    if (wert & 0x80) {
        if (wert & 0x40) {
	    len += sprintf (buf + len, "Radio Controlled Adjusted\n");
	}
	else  len += sprintf (buf + len, "Radio Controlled\n");
    }
    else if (wert & 0x40)
        len += sprintf (buf + len, "Crystal Operation\n");
    else  len += sprintf (buf + len, "Time Invalid\n");
    /* get quality */
    wert = inb_p (port_address + GUETE_OFFSET);
    wert = wert / 2.55 + .5;
    len += sprintf (buf + len, "Quality: %d\%\n", wert);

    return len;
}            

static struct proc_dir_entry hopf_proc_entry =
{
    0,                          /* low_ino: the inode -- dynamic */
    5, "dcf77",                   /* len of name and name */
    S_IFREG | S_IRUGO,          /* mode */
    1, 0, 0,                    /* nlinks, owner, group */
    0, NULL,                    /* size - unused; operations -- use default */
    &hopf_read_proc,          /* function used to read data */
    /* nothing more */
};    


/******************************************************************************
 *
 * Name        : hopf_open
 * Description : 
 * Input       : -
 * Output      : -
 *
 *****************************************************************************/
static int hopf_open(struct inode *inode, struct file *filp)
{
    DPRINTK("hopf_open\n");
    
    filp->f_pos = 0;

    MOD_INC_USE_COUNT;
    return 0;                   /* success */
}


/******************************************************************************
 *
 * Name        : hopf_release
 * Description : 
 * Input       : -
 * Output      : -
 *
 *****************************************************************************/
static int hopf_release(struct inode *inode, struct file *filp)
{
    DPRINTK("hopf_release\n");

    MOD_DEC_USE_COUNT;
    return 0;
}

/******************************************************************************
 *
 * read, write, ioctl, ...
 *
 * not filled in yet, cause there is a little lack of knowledge, but I'm 
 * working hard to fit the holes!
 *
 *****************************************************************************/

/******************************************************************************
 *
 * Name        : hopf_read
 * Description : 
 * Input       : -
 * Output      : -
 *
 *****************************************************************************/
static ssize_t hopf_read(struct file *filp, char *buf, 
			    size_t count, loff_t *ppos)
{
    int result;
    unsigned char wert, byte;
    unsigned int msec;
    unsigned char kbuf[12];

    DPRINTK("getting time from DCF77-Card\n");

    /* get hour */
    wert = inb_p (port_address + 2);
    byte = wert >> 4;
    kbuf[0] = byte + 48;
    byte = wert & 0xf;
    kbuf[1] = byte + 48;
    /* get minutes */
    kbuf[2] = ':';
    wert = inb_p (port_address + 1);
    byte = wert >> 4;
    kbuf[3] = byte + 48;
    byte = wert & 0xf;
    kbuf[4] = byte + 48;
    /* get seconds */
    kbuf[5] = ':';
    wert = inb_p (port_address + 0);
    byte = wert >> 4;
    kbuf[6] = byte + 48;
    byte = wert & 0xf;
    kbuf[7] = byte + 48;
    /* get millisconds */
    kbuf[8] = ':';
    msec = inw (port_address + 0x38);
    byte = msec / 100;
    kbuf[9] = byte + 48;
    msec -= byte * 100;
    byte = msec / 10;
    kbuf[10] = byte + 48;
    msec -= byte * 10;
    kbuf[11] = msec + 48;

    /* copy string to user space */
    result = copy_to_user (buf, kbuf, 12);

/*     kfree(kbuf); */
    return result;
}


void GetHardwareTime(HOPFTIME *Data,unsigned int Offset)
{
    int i;
    u32 value;
    memset(Data, 0, sizeof(HOPFTIME));

    Data->wMilliseconds=inw (port_address + MSEC_OFFSET);
    if( Offset == 16 ) {
	    for (i=0; i<HOPF_TIMESIZE; i++) {
	       hopf_time  [i] = inb (port_address + UTC_OFFSET + i);
	    }
    } else {
	    for (i=0; i<HOPF_TIMESIZE; i++) {
	       hopf_time  [i] = inb (port_address + MEZ_OFFSET + i);
	    }
    }
    Data->wYear = hopf_time[6]+2000;
    Data->wMonth= hopf_time[5];
    Data->wDayOfWeek=hopf_time[4];
    Data->wDay  = hopf_time[3];
    Data->wHour = hopf_time[2];
    Data->wMinute=hopf_time[1];
    Data->wSecond=hopf_time[0];
    Data->wStatus =0;

    return;
}



/******************************************************************************
 *
 * Name        : hopf_ioctl
 * Description : 
 * Input       : arg - should always be a pointer to the related 
 *                     data-structure. The size of the structure can be
 *                     found out by using the size macros.
 * Output      : -
 *
 *****************************************************************************/
static int hopf_ioctl(struct inode *inode, struct file *filp,
                  unsigned int cmd, unsigned long arg)
{
    unsigned int msec;
    unsigned long ul_error;
    int i;
    int size;
/*     int result; */
    unsigned char guete;

//#if 0
    size = _IOC_SIZE(cmd);  /* the size bitfield in cmd */
    /*
     * extract the type and number of bitfields, and don't decode
     * wrong cmds: return -EINVAL before verify_area().
     */
//    if (_IOC_TYPE(cmd) != HOPF_IOC_MAGIC) {
//        PERRORK("you didn't use the correct magic number!\n");
//        return -EINVAL;
//    }
//    if (_IOC_NR(cmd) > HOPF_IOC_MAXNR) {
//        PERRORK("you didn't use a valid magic ioctl number!\n");
//        return -EINVAL;
//    }
    
    switch (cmd) {

    case HOPF_IOCS_RESET:
        DPRINTK("HOPF_IOCS_RESET\n");
	outb_p (0x01, port_address + COMMAND_OFFSET);
	break;
    case HOPF_IOCS_MASTER:
        DPRINTK("HOPF_IOCS_MASTER\n");
	outb (0x00, port_address + COMMAND_OFFSET);
	outb (0x00, port_address + SET_OFFSET);
	outb (0x32, port_address + COMMAND_OFFSET);
        break;
    case HOPF_IOCS_SLAVE:
        DPRINTK("HOPF_IOCS_SLAVE\n");
	outb (0x00, port_address + COMMAND_OFFSET);
	outb (0x80, port_address + SET_OFFSET);
	outb (0x32, port_address + COMMAND_OFFSET);
        break;
    case HOPF_IOCG_STATUS:
        DPRINTK ("HOPF_IOCG_STATUS\n");
	hopf_status = inb_p (port_address + 0x07);
	PINFOK ("Status = %d\n", hopf_status);
	put_user (hopf_status, (unsigned char *) arg);
	break;
    case HOPF_IOCG_TIME:
        DPRINTK ("HOPF_IOCG_TIME\n");
	for (i=0; i<HOPF_TIMESIZE; i++) {
	    hopf_time[i] = inb_p (port_address + MEZ_OFFSET + i);
	}
	ul_error = copy_to_user ((unsigned char *) arg, hopf_time,
				 HOPF_TIMESIZE);
	if (ul_error < 0)  return ((int) ul_error);
	else {
	    msec = inw (port_address + MSEC_OFFSET);
	    return ((int) msec);
	}
	break;
    case HOPF_IOCS_SET:
        DPRINTK ("HOPF_IOCS_SET\n");
	ul_error = copy_from_user ((unsigned char *) arg, hopf_time,
				   HOPF_TIMESIZE);
	if (ul_error < 0)  return ((int) ul_error);
	    outb_p (0x56, port_address + SET_OFFSET + 0);
	    outb_p (0x34, port_address + SET_OFFSET + 1);
	    outb_p (0x12, port_address + SET_OFFSET + 2);
	    outb_p (0x03, port_address + SET_OFFSET + 3);
	    outb_p (0x23, port_address + SET_OFFSET + 4);
	    outb_p (0x04, port_address + SET_OFFSET + 5);
	    outb_p (0x01, port_address + SET_OFFSET + 6);
	    outb_p (0x50, port_address + SET_OFFSET + 7);
	PERRORK("Unable to get major %x\n", hopf_time[0]);
	PERRORK("Unable to get major %x\n", hopf_time[1]);
	PERRORK("Unable to get major %x\n", hopf_time[2]);
	PERRORK("Unable to get major %x\n", hopf_time[3]);
	PERRORK("Unable to get major %x\n", hopf_time[4]);
	PERRORK("Unable to get major %x\n", hopf_time[5]);
	PERRORK("Unable to get major %x\n", hopf_time[6]);
//	for (i=0; i<HOPF_TIMESIZE; i++) {
//	    outb (hopf_time[i], port_address + SET_OFFSET + i);
//	}
	break;
    case HOPF_IOCS_PORT:
        DPRINTK ("HOPF_IOCS_PORT\n");
	if ((arg  < 0x200) || (arg > 0x3f0))
	    return (-EFAULT);
/* 	outb_p ((unsigned char) arg>>4, port_address + SET_OFFSET); */
/* 	outb_p (0x14, port_address + COMMAND_OFFSET); */
/* 	PINFOK ("setting port address to 0x%x\n", (unsigned int) arg); */
/* 	release_region(port_address, PORT_SIZE); */
/* 	port_address = arg; */
/* 	if ((result = check_region(port_address, PORT_SIZE)) < 0) { */
/* 	    DPRINTK("port is busy!\n"); */
/* 	    unregister_chrdev(HOPF_MAJOR, DEV_NAME); */
/* 	    return result; */
/* 	} */
/* 	request_region(port_address, PORT_SIZE, DEV_NAME);  */
	break;
    case HOPF_IOCQ_PORT:
        DPRINTK ("HOPF_IOCQ_PORT\n");
	ul_error = hopf_test ();
	return (ul_error);
    case HOPF_IOCQ_GUETE:
        DPRINTK ("HOPF_IOCQ_GUETE\n");
	guete = inb_p (port_address + GUETE_OFFSET);
	return (guete);
	break;
//    default:
//       PERRORK("unknown ioctl request\n");
//        return -EINVAL;
        /* if we want to be POSIX compatible we have to use -ENOTTY here */
    }
//#else    
  /* process ioctl-command */
  switch(cmd & HOPF_CLOCK_CMD_MASK) {

  case HOPF_CLOCK_GET_UTC:   // must be the first entry
    {
        HOPFTIME Data;
        GetHardwareTime(&Data,16);
        copy_to_user((HOPFTIME*)arg, &Data,sizeof(HOPFTIME));
    };
    break;

  case HOPF_CLOCK_GET_LOCAL:
    {
        HOPFTIME Data;
        GetHardwareTime(&Data,0);
        copy_to_user((HOPFTIME*)arg,
            &Data,sizeof(HOPFTIME));
    };
    break;
    
  case HOPF_CLOCK_SET_POSITION:
  {
  	DPRINTK("HOPF_SET_POSITON\n");
  }
    break;
 
  case HOPF_CLOCK_SET_TIME:
    break;

  case HOPF_CLOCK_SET_ANTENNA:
    break;

  case HOPF_CLOCK_SET_DIFFERENCE:
  {
  	DPRINTK("HOPF_SET_DIFFERENCE\n");
  }
    break;

  case HOPF_CLOCK_SET_SATMODE:
    break;

  case HOPF_CLOCK_SET_RESET:
  {
  	DPRINTK("HOPF_SET_RESET\n");
  }
    break;

  case HOPF_CLOCK_SET_SYSTEMBYTE:
    break;

  case HOPF_CLOCK_SET_IRIG:
    break;

  case HOPF_CLOCK_GET_IRIG:
    break;

  case HOPF_CLOCK_GET_SYSTEMBYTE:
    break;

  case HOPF_CLOCK_GET_SATDATA:
    break;

  case HOPF_CLOCK_GET_POSITION:
  {
  	DPRINTK("HOPF_GET_POSITION\n");
  }
    break;

  case HOPF_CLOCK_GET_ANTENNA:
    break;

  case HOPF_CLOCK_GET_VERSION:
  {
	ClockVersion Data;
	strcpy(Data.cVersion, "DCF-RECEIVER");
	strcpy(Data.dVersion, "0.1");
	copy_to_user((ClockVersion*)arg, &Data,sizeof(ClockVersion));
  }
    break;

 case HOPF_CLOCK_GET_DIFFERENCE:
 {
 	DPRINTK("HOPF_GET_DIFFERENCE\n");
 }
    break;
  };
//#endif
    return 0;
}


static struct file_operations hopf_fops =
{
    NULL,
    NULL,                   /* lseek */
    hopf_read,              /* read */
    NULL,                   /* write */
    NULL,                   /* readdir */
    NULL,                   /* poll? select? */
    hopf_ioctl,             /* ioctl */
    NULL,                   /* mmap */
    hopf_open,              /* open */
    NULL,                   /* flush */       
    hopf_release,           /* release */
    NULL,                   /* fsync */
    NULL,                   /* fasync */
    NULL,                   /* check_media_change */
    NULL                    /* revalidate */
};


/******************************************************************************
 *
 * Name        : init_module
 * Description : 
 * Input       : -
 * Output      : on error -EIO unable to get major
 *               otherwise 0 if erverything was ok.
 *
 *****************************************************************************/
int init_module(void)
{
    int result = -ENODEV;

    PINFOK("initialization ...\n");
 
    result = register_chrdev(hopf_major, DEV_NAME, &hopf_fops);
    if (result < 0) {
        PERRORK("Unable to get major %d\n", HOPF_MAJOR);
        return result;
    }
    if (hopf_major == 0)
        hopf_major = result;

 /* Testing /proc/ioports */
    result = hopf_test ();
    if (result < 0) {
	PERRORK ("No device found!\n");
	unregister_chrdev(HOPF_MAJOR, DEV_NAME);
	return result;
    }
    PINFOK ("hopf DCF77-card found at 0x%x.\n", (unsigned int) port_address);
    /*
     * request IO-ports
     */
    if ((result = check_region(port_address, PORT_SIZE)) < 0) {
        DPRINTK("port is busy!\n");
	unregister_chrdev(HOPF_MAJOR, DEV_NAME);
        return result;                                      /* busy      */
    }
    request_region (port_address, PORT_SIZE, DEV_NAME); 
//    proc_register (&proc_root, &hopf_proc_entry);

    PINFOK("-------------------------------------------\n");
    PINFOK("(-: hopf driver successfully installed!! \n");
    PINFOK("                 Version " VERSION "\n");
    PINFOK("-------------------------------------------\n");

    return 0;
}


/******************************************************************************
 *
 * Name        : cleanup_module
 * Description : Free kernelmemory of the SDBs and free the irq used by the
 *               cards. Afterwards unregister the device.
 * Input       : -
 * Output      : -
 *
 *****************************************************************************/
void cleanup_module(void)
{
    PINFOK("cleanup & unregister ...\n");

//    proc_unregister (&proc_root, hopf_proc_entry.low_ino);
    release_region(port_address, PORT_SIZE);

    unregister_chrdev(HOPF_MAJOR, DEV_NAME);

    PINFOK("-------------------------------------------\n");
    PINFOK(DEV_NAME " ;-)   cleanup & unregister ok!!   \n");
    PINFOK("                 Version " VERSION "\n");
    PINFOK("-------------------------------------------\n");
}
